import getpass
import json
import os

import geopandas as gpd
from elogs.s3 import S3BucketReader
from loguru import logger
from rasterio.session import AWSSession
from smart_open import open

CDSE_ENDPOINT_URL = "https://eodata.dataspace.copernicus.eu/"
VEGAWS_BUCKET = "vito-lcfm"
SENTHUB_BUCKET_NAME = "sentinel-s2-l2a"
VEGAWS_OUTPUT_BUCKETNAME = "vito-lcfm"

USERNAME = getpass.getuser()
CDSE_BUCKET_NAME = "EODATA"
COLLECTION_S3_URI = "s3://vito-lcfm/global_processing/collections/lcfm-10_collection_baseV2_senthub_2020-01-01_2020-12-31.csv"
RIO_CONFIG_URI = "s3://vito-lcfm/global_processing/config/rio_config.json"
FGB_10_S3_URI = "s3://vito-lcfm/global_processing/collections/LCFM_10p_S2-tiles.fgb"
FGB_100_S3_URI = "s3://vito-lcfm/global_processing/collections/LCFM_100p_S2-tiles.fgb"
MODEL_INDEXES = "s3://vito-lcfm/global_processing/models/model_indexes.json"

band_names = {
    60: [
        "B01",
        "B02",
        "B03",
        "B04",
        "B05",
        "B06",
        "B07",
        "B8A",
        "B09",
        "B11",
        "B12",
        "SCL",
    ]
}


def get_shapefiles_tilelist(fgb_path):
    shapefiles_path = os.path.join(os.path.dirname(__file__), "shapefiles")
    gdf100 = gpd.read_file(f"{shapefiles_path}/LCFM_100p_S2-tiles.fgb")
    return gdf100["tile"].tolist()


class UTMTile:
    def __init__(self, epsg, tile, bounds, geometry):
        self.epsg = epsg
        self.tile = tile
        self.bounds = bounds
        self.geometry = geometry

    def get_geometry(self):
        return self.geometry

    @classmethod
    def from_geodataframe(cls, tile_value, fgb_path: str):
        with open(fgb_path, "r") as f:
            fgb_file = f.read()
        gdf = gpd.read_file(fgb_file)
        row = gdf[gdf["tile"] == tile_value].iloc[0]
        epsg = row["epsg"]
        tile = row["tile"]
        bounds = eval(
            row["bounds"]
        )  # Assuming 'bounds' is a string representation of a bounding box
        geometry = row["geometry"]
        return cls(epsg, tile, bounds, geometry)


def get_cdse_awssession():
    """There is a peculiarity in rasterio env not being able to read https url
    check the comments in issue: https://github.com/rasterio/rasterio/issues/3152"""
    # cdse_env = dotenv_values(os.path.expanduser(CDSE_CREDENTIALS))
    session = AWSSession(
        aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        endpoint_url=CDSE_ENDPOINT_URL,
        # region_name=os.getenv("AWS_DEFAULT_REGION")
    )
    return session


def get_key_from_s3_uri(s3_uri: str) -> str:
    """
    Extracts the key from a full S3 URI path.

    :param s3_uri: The full S3 URI (e.g., 's3://bucket-name/path/to/object')
    :return: The key (e.g., 'path/to/object')
    """
    if not s3_uri.startswith("s3://"):
        raise ValueError("Invalid S3 URI. It must start with 's3://'")

    return "/".join(s3_uri.split("/")[3:])


def download_checkpoint(checkpoint, model_name, tgt_path) -> str:
    """Downloads and returns the tgt path of the collection models"""
    "s3://vito-lcfm/global_processing/models/2412101640_v4test5/version_0/checkpoints"

    PREFIX = f"global_processing/models/{model_name}/version_0/checkpoints"
    # checkpoint_key = get_key_from_s3_uri(checkpoint)
    s3_api = get_s3_api(VEGAWS_BUCKET)
    s3_api.download(
        PREFIX + "/" + checkpoint, tgt_path, verbose=True
    )  # NOTE Can the checkpoint read on RAM?

    return tgt_path


def _init_s3_api(bucket):
    return S3BucketReader.from_credentials(
        aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        bucket=bucket,
        aws_region="eu-central-1",
    )


def get_s3_api(bucket) -> S3BucketReader:
    return _init_s3_api(bucket=bucket)


def get_vegaws_session():
    """There is a peculiarity in rasterio env not being able to read https url
    check the comments in issue: https://github.com/rasterio/rasterio/issues/3152"""
    session = AWSSession(
        aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        region_name="eu-central-1",
    )
    return session


class LOIConfig:
    """Creates a config object for cloud model inference"""

    def __init__(self, config_path: str):
        self.config_path = config_path
        args = self.__load_config(config_path)

        self.model_name = args["model_name"]
        self.model_idx = self.__get_model_index(self.model_name)
        self.version_export = args["version_export"]
        self.collection_uri = args["collection_uri"]
        self.overwrite = args.get("overwrite", False)
        self.skip_done = args.get("skip_done", True)
        self.testing = args.get("testing", False)
        self.local = args.get("local", False)
        self.spark_slices_factor = args.get("spark_slices_factor", 1)
        self.rio_config_path = args.get("rio_config_path", RIO_CONFIG_URI)

    def __repr__(self):
        # Create a dictionary excluding self.s3_vegaws
        repr_dict = {k: v for k, v in self.__dict__.items() if k != "s3_vegaws"}
        return json.dumps(repr_dict, indent=4)

    @staticmethod
    def __load_config(_config_path: str):
        s3_vegaws_api = get_s3_api(bucket=VEGAWS_BUCKET)
        if _config_path.startswith("s3://"):
            conf_json = get_key_from_s3_uri(_config_path)
            args: dict = s3_vegaws_api.read_json(conf_json)
        else:
            conf_json = _config_path
            with open(conf_json, 'r') as f:
                args = json.load(f)
        logger.info(f"Reading configuration from {conf_json}")
        return args

    @staticmethod
    def __get_model_index(model_name):
        s3_vegaws_api = get_s3_api(bucket=VEGAWS_BUCKET)
        model_idx = s3_vegaws_api.read_json(get_key_from_s3_uri(MODEL_INDEXES))[
            "model_name"
        ][model_name]
        logger.info(f"Model name: {model_name}, Model index: {model_idx}")
        return model_idx
