"""
Cloud mask inference pipeline for LOI-60 products.
Contact: YK
"""

import json
import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import geopandas as gpd
import numpy as np
import rasterio
from evotrain.models.loi.data.bands import (
    build_bands_data_array,
    get_filenames_dict,
)
from evotrain.models.loi.data.preprocessing import scale_s2_l2a_bands
from evotrain.models.loi.infer import get_feats_head, load_model_files
from evotrain.models.loi.io import save_geotiff
from loguru import logger
from smart_open import open

from veg_workflows.collections import veg_collections
from veg_workflows.paths.models import CLOUDSEN_MODELS_60M
from veg_workflows.products import Loi10Product


class CloudInferencePipeline:
    def __init__(
        self,
        model_name: Optional[str] = None,
        model_idx: Optional[int] = -1,
        version: Optional[str] = "v999",
        export_products: bool = False,
        host: str = "terrascope",  # either "terrascope" or "aws"
        # let's add aws_config as a path pointing to the config file
        aws_config: Optional = None,
    ):
        """Initialize the cloud inference model."""

        # Note that this is like a "global" settings for the pipeline
        # but it can be overridden in the run_inference method if needed
        self.export_products = export_products

        # Set the host
        self.host = host

        if self.host not in ["terrascope", "aws"]:
            raise ValueError("host must be either 'terrascope' or 'aws'")
        logger.info(f"Host set to: {host}")

        # Load model
        if self.host == "terrascope":
            if not model_name:
                raise ValueError(
                    "model_name must be provided when host is 'terrascope'"
                )
            if model_idx is None or model_idx < -1:
                raise ValueError(
                    "model_idx must be provided and be >= -1 when host is 'terrascope'"
                )

            self.model_name = model_name
            self.model_idx = model_idx

            self.version = version
            # # Set torch hub directory
            # username = getpass.getuser()
            # torch.hub.set_dir(f"/data/users/Public/{username}/share/.torchhub_cache")

            self.load_model(self.model_name, self.model_idx)
        elif self.host == "aws":
            from veg_workflows.aws.global_processing import LOIConfig

            self.aws_config = aws_config
            if not isinstance(self.aws_config, LOIConfig):
                raise ValueError("aws_config must be a LOIConfig object")

            self.model_name = self.aws_config.model_name
            self.model_idx = self.aws_config.model_idx
            self.version = self.aws_config.version_export

            self.rio_config_path = self.aws_config.rio_config_path
            try:
                with open(self.rio_config_path, "r") as rio_config:
                    self.rio_config = json.loads(rio_config.read())
            except Exception as e:
                # Let's try in the same directory as the config file
                self.rio_config_path = (
                    Path(self.aws_config.config_path).parent
                    / "rio_config.json"
                )
                with open(self.rio_config_path, "r") as rio_config:
                    self.rio_config = json.loads(rio_config.read())
            logger.info(f"Loaded rio config from {self.rio_config_path}")

            self.load_model_aws(self.model_name, self.model_idx)

        logger.info("Model initialized successfully")

    def load_model(self, model_name: str, model_idx: int = -1):
        """Load the model from the given path."""

        self.model_path = f"{CLOUDSEN_MODELS_60M}/{self.model_name}/version_0/"
        self.model, self.class_mapping, self.model_config = load_model_files(
            self.model_path, idx=model_idx, load_last=False
        )

    def load_model_aws(
        self,
        model_name: str,
        model_idx: int = -1,
        vegaws_bucket: str = "vito-lcfm",
    ):
        from evotrain.models.loi.infer.load import load_s3_model_files

        self.model_path = f"s3://{vegaws_bucket}/global_processing/models/{self.model_name}/version_0"

        self.model, self.class_mapping, self.model_config = (
            load_s3_model_files(
                self.model_path, idx=self.model_idx, load_last=False
            )
        )

    def get_input_aws(
        self,
        product_id: str,
        product_base_path: Path,
        product_name: str,
        sentinelhub_bucket: str,
        s3_uri: str,
        s2grid_uri: str,
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        Get and prepare input data from a product ID using AWS resources.
        """

        from veg_workflows.aws.global_processing import (
            band_names,
            get_key_from_s3_uri,
            get_s3_api,
            get_vegaws_session,
        )

        loi10prod, tile, epsg, bounds, date = (
            prepare_loi_product_for_inference_aws(
                s2_product_id=product_id,
                version=self.version,
                product_base_path=product_base_path,
                product_name=product_name,
                s2grid_uri=s2grid_uri,
            )
        )

        logger.info(f"Processing product: {product_id}")
        s3_senthub = get_s3_api(bucket=sentinelhub_bucket)

        prefix = get_key_from_s3_uri(s3_uri)
        band_keys = s3_senthub.list_files(
            prefix=prefix + "R60m/",
        )

        # keep only band keys that contain the band names
        band_keys = [
            band_key
            for band_key in band_keys
            if any(band_name in band_key for band_name in band_names[60])
        ]
        # s3_bandpaths = [s3_safe_uri +'/'+ os.path.basename(band_key) for band_key in band_keys]
        s3_bandpaths = [
            f"s3://{sentinelhub_bucket}/{band_key}" for band_key in band_keys
        ]

        filenames_dict = {
            60: {
                key: band_path
                for band_path in s3_bandpaths
                for key in [os.path.splitext(os.path.basename(band_path))[0]]
            }
        }

        bands_list = self.model_config["bands_config"]["s2_bands"]
        bands_list = [
            band[4:] if band.startswith("l2a-") else band
            for band in bands_list
        ]

        if os.path.exists(self.rio_config["CPL_LOG"]):
            os.remove(self.rio_config["CPL_LOG"])

        with rasterio.Env(session=get_vegaws_session(), **self.rio_config):
            darr_og = build_bands_data_array(
                filenames_dict,
                bands_list,
                source_resolution=60,
                target_resolution=60,
                apply_offset=False,
            )

            scl_array = build_bands_data_array(
                filenames_dict,
                bands_list=["SCL"],
                source_resolution=60,
                target_resolution=60,
            )

        scl_array = scl_array.sel(band="SCL").values

        metadata = {
            "loi10prod": loi10prod,
            "epsg": epsg,
            "bounds": bounds,
            "date": date,
            "tile": tile,
            "bands": bands_list,
            "scl_array": scl_array,
        }

        self.current_metadata = metadata

        return darr_og, metadata

    def get_input(
        self, product_id: str, collection_name: str
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """Get and prepare input data from a product ID.

        Args:
            product_id: Sentinel-2 product ID
            collection_name: Name of the collection

        Returns:
            Tuple containing the input array and metadata
        """
        logger.info(f"Processing product: {product_id}")

        # Prepare product
        loi10prod, product_collection, epsg, bounds, date = (
            prepare_loi_product_for_inference(
                product_id, collection_name, self.version
            )
        )

        # Get bands data
        bands_list = self.model_config["bands_config"]["s2_bands"]
        bands_list = [
            band[4:] if band.startswith("l2a-") else band
            for band in bands_list
        ]
        filenames_dict = get_filenames_dict(product_collection, bands_list)
        darr_og = build_bands_data_array(
            filenames_dict,
            bands_list,
            source_resolution=60,
            target_resolution=60,
        )

        # Get SCL band
        filenames_dict = get_filenames_dict(
            product_collection, ["SCL"], source_resolution=60
        )
        scl_array = (
            build_bands_data_array(
                filenames_dict,
                bands_list=["SCL"],
                source_resolution=60,
                target_resolution=60,
            )
            .sel(band="SCL")
            .values
        )

        metadata = {
            "loi10prod": loi10prod,
            "epsg": epsg,
            "bounds": bounds,
            "date": date,
            "bands": bands_list,
            "scl_array": scl_array,
        }

        self.current_metadata = metadata

        return darr_og, metadata

    def preprocess(
        self, input_array: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Preprocess the input data.

        Args:
            input_array: Raw input array
        Returns:
            Tuple of preprocessed data and features
        """
        if not hasattr(self, "current_metadata"):
            raise RuntimeError("get_input must be called before preprocess")

        # Apply SCL mask
        scl_array = self.current_metadata["scl_array"]
        input_array = input_array.where(scl_array != 0, np.nan)
        if np.isnan(input_array).any(axis=0).data.any():
            logger.debug("Block has NaNs")
            input_array = input_array.fillna(0)

        # Scale data
        scaled_data = scale_s2_l2a_bands(
            input_array.data,
            scaling_factor=10_000,
            apply_logistic=True,
            k_factor=self.model_config["scaling_config"]["k_factor"],
            apply_jitter=False,
            k_factor_jitter=0,
        )

        # Generate additional features
        feats_head = get_feats_head(
            scaled_data.shape,
            self.current_metadata["epsg"],
            self.current_metadata["bounds"],
            self.current_metadata["date"],
            self.model_config,
        )

        return scaled_data, feats_head

    def preprocess_aws(
        self, input_array: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        return self.preprocess(input_array)

    def run_inference(
        self,
        preprocessed_data: np.ndarray,
        feats_head: np.ndarray,
        export: bool = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Run model inference on preprocessed data.

        Args:
            preprocessed_data: Preprocessed input data
            feats_head: Additional features
            export: Whether to export the results

        Returns:
            Tuple of probability outputs and class outputs (argmax)
        """
        should_export = export if export is not None else self.export_products

        logger.info("Starting model inference")
        model_output = self.model.predict(preprocessed_data, feats_head)
        out = model_output.squeeze()

        # Get argmax output
        out_argmax = np.argmax(out.copy(), axis=0).astype(np.float32)

        # Apply masks
        scl_array = self.current_metadata["scl_array"]
        mask_scl_invalid = np.where(scl_array == 0)
        out[:, mask_scl_invalid[0], mask_scl_invalid[1]] = np.nan

        # Process outputs
        mask_out_invalid = np.isnan(out)
        nodata_value = 255
        out = np.round(out * 250).astype(np.uint8)
        out[mask_out_invalid] = nodata_value
        out_argmax[mask_scl_invalid] = nodata_value

        if should_export:
            if self.host == "terrascope":
                self.export_results(out, out_argmax, nodata_value)
            elif self.host == "aws":
                self.export_results_aws(out, out_argmax, nodata_value)
            else:
                raise ValueError("host must be either 'terrascope' or 'aws'")
            logger.success("Results exported successfully")

        logger.success("Model inference completed successfully")

        return out, out_argmax

    def run_inference_aws(
        self,
        preprocessed_data: np.ndarray,
        feats_head: np.ndarray,
        export: bool = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        return self.run_inference(preprocessed_data, feats_head, export)

    def export_results(
        self,
        prob_output: np.ndarray,
        class_output: np.ndarray,
        nodata_value: int,
        prob_path: Optional[Path] = None,
        mask_path: Optional[Path] = None,
    ):
        """Export results to GeoTIFF files.

        Args:
            prob_output: Probability outputs to save
            class_output: Class outputs to save
            nodata_value: Value to use for no data
            prob_path: Optional custom path for probability file
            mask_path: Optional custom path for mask file
        """
        if not self.model_config["data_config"]["classify_snow"]:
            raise NotImplementedError(
                "Only snow classification is supported for now"
            )

        bands_names = [
            f"PROB_{k}"
            for k in self.model_config["labels_config"][
                "cloudsen12_mergedclouds_extrasnow"
            ].keys()
        ]
        tags = {
            "bands_names": bands_names,
            "legend": "0 Surface \n 1 Clouds \n2 Shadows \n3 Snow \n255 No data",
        }

        # Use custom path if provided, otherwise use default from loi10prod
        save_path_probs = prob_path or self.current_metadata["loi10prod"].path(
            "probs"
        )
        save_path_mask = mask_path or self.current_metadata["loi10prod"].path(
            "mask"
        )

        # Save probability output
        save_geotiff(
            prob_output,
            self.current_metadata["bounds"],
            self.current_metadata["epsg"],
            save_path_probs,  # Use the chosen path
            scales=[1 / 250] * prob_output.shape[0],
            nodata_value=nodata_value,
            bands_names=bands_names,
            tags=tags,
        )

        # Save class output
        save_geotiff(
            class_output.astype(np.uint8),
            self.current_metadata["bounds"],
            self.current_metadata["epsg"],
            save_path_mask,  # Use the chosen path
            nodata_value=nodata_value,
            bands_names=["CLASS"],
            tags={"band_names": ["CLASS"], "legend": tags["legend"]},
        )

    def export_results_aws(
        self,
        prob_output: np.ndarray,
        class_output: np.ndarray,
        nodata_value: int,
        prob_path: Optional[Path] = None,
        mask_path: Optional[Path] = None,
    ):
        """Export results to GeoTIFF files using AWS resources."""

        from veg_workflows.aws.global_processing import get_vegaws_session

        with rasterio.Env(
            session=get_vegaws_session(),
            CPL_VSIL_USE_TEMP_FILE_FOR_RANDOM_WRITE="YES",
            CPL_DEBUG=self.rio_config.get("CPL_DEBUG"),
            CPL_CURL_VERBOSE=self.rio_config.get("CPL_CURL_VERBOSE"),
            CPL_LOG=self.rio_config.get("CPL_LOG"),
            CPL_LOG_ERRORS=self.rio_config.get("CPL_LOG_ERRORS"),
        ):
            if not self.model_config["data_config"]["classify_snow"]:
                raise NotImplementedError(
                    "Only snow classification is supported for now"
                )

            bands_names = list(
                self.model_config["labels_config"][
                    "cloudsen12_mergedclouds_extrasnow"
                ].keys()
            )
            bands_names = [f"PROB_{k}" for k in bands_names]
            bands_tags = [
                {
                    f"PROB_{k}_CHANNEL": v
                    for k, v in self.model_config["labels_config"][
                        "cloudsen12_mergedclouds_extrasnow"
                    ].items()
                }
                for _ in range(prob_output.shape[0])
            ]
            tags = {
                "bands_names": bands_names,
                "legend": "0 Surface \n 1 Clouds \n2 Shadows \n3 Snow \n255 No data",
            }

            # Use custom path if provided, otherwise use default from loi10prod
            save_path_probs = prob_path or self.current_metadata[
                "loi10prod"
            ].path("probs")
            save_path_mask = mask_path or self.current_metadata[
                "loi10prod"
            ].path("mask")

            # Save the probabilities output
            save_geotiff(
                prob_output,
                self.current_metadata["bounds"],
                self.current_metadata["epsg"],
                save_path_probs,
                scales=[1 / 250] * prob_output.shape[0],
                nodata_value=nodata_value,
                bands_names=bands_names,
                tags=tags,
            )
            logger.success(
                f"Saved the probabilities output to {save_path_probs}"
            )

            # Save the argmax output
            bands_tags = [bands_tags[0]]
            bands_names = [
                k[5:] for k in bands_names
            ]  # Removing the "PROB_" prefix
            bands_tags = [
                {
                    k.replace("_CHANNEL", ""): v
                    for k, v in bands_tags[0].items()
                }
            ]
            tags = {
                "band_names": ["CLASS"],
                "legend": "0 Surface \n 1 Clouds \n2 Shadows \n3 Snow \n255 No data",
            }
            save_geotiff(
                class_output.astype(np.uint8),
                self.current_metadata["bounds"],
                self.current_metadata["epsg"],
                save_path_mask,
                nodata_value=nodata_value,
                bands_names=["CLASS"],
                tags={"band_names": ["CLASS"], "legend": tags["legend"]},
            )
            logger.success(f"Saved the argmax output to {save_path_mask}")

    def process(
        self, product_id: str, collection_name: str, export: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Process a product end-to-end in one call."""
        input_array, _ = self.get_input(product_id, collection_name)
        preprocessed_data, feats_head = self.preprocess(input_array)
        return self.run_inference(preprocessed_data, feats_head, export)


def prepare_loi_product_for_inference_aws(
    s2_product_id: str,
    version: str,
    product_base_path: Path,
    product_name: str,
    s2grid_uri: str,
) -> Tuple[Loi10Product, object, int, tuple, str]:
    # Create the LOI-10 product
    loi10prod = create_loi10_product_aws(
        s2_product_id=s2_product_id,
        project_name="LCFM",
        version=version,
        products_base_path=product_base_path,
        product_name=product_name,
    )

    with open(s2grid_uri, "rb") as s2grid_rb:
        s2grid = gpd.read_file(s2grid_rb)
        s2grid["bounds"] = s2grid["bounds"].apply(eval)

    tile = s2_product_id.split("_")[5][1:]
    bounds = s2grid[s2grid.tile == tile].iloc[0].bounds

    date_id = loi10prod.date_id.split("_")[0]
    date_id = datetime.strptime(date_id, "%Y%m%dT%H%M%S")
    date = date_id.strftime("%Y-%m-%d")
    epsg = loi10prod.epsg

    return loi10prod, tile, epsg, bounds, date


def prepare_loi_product_for_inference(
    s2_product_id: str, collection_name: str, version: str
) -> Tuple[Loi10Product, object, int, tuple, str]:
    """
    Prepare LOI product by loading its information and creating necessary directories.

    Args:
        s2_product_id (str): Sentinel-2 product ID.
        collection_name (str): Name of the collection from veg_collections (e.g. "lcfm_1percent", "lcfm_10percent").
        version (str): Version of the product (e.g. "v999").

    Returns:
        Tuple[Loi10Product, object, int, tuple, str]: A tuple containing the LOI product, product collection, EPSG code, bounds, and date.
    """
    logger.success("Preparing LOI product")

    # Create the LOI-10 product
    loi10prod = create_loi10_product(
        s2_product_id,
        "LCFM",
        version,
        Path("/vitodata/vegteam_vol2/products/"),
        "LOI-10",
    )

    # Load the collection and s2grid
    veg_collection, s2grid = get_collection_and_s2grid(collection_name)

    # Filter the product collection
    product_collection = filter_product_collection(
        veg_collection, s2grid, s2_product_id
    )

    # Extract product information
    bounds, date, epsg, product_tile = extract_product_info(
        product_collection, s2grid
    )

    return loi10prod, product_collection, epsg, bounds, date


def create_loi10_product(
    s2_product_id: str,
    project_name: str,
    version: str,
    products_base_path: Path,
    product_name: str,
) -> Loi10Product:
    """
    Create a Loi10Product object and the necessary directories.

    Args:
        s2_product_id (str): Sentinel-2 product ID.
        project_name (str): Name of the project.
        version (str): Version of the product.
        products_base_path (Path): Base path for the products.
        product_name (str): Name of the product.

    Returns:
        Loi10Product: A Loi10Product object.
    """
    loi10prod = Loi10Product(
        s2_product_id=s2_product_id,
        project_name=project_name,
        version=version,
        products_base_path=str(products_base_path),
        product_name=product_name,
    )

    # Create the necessary directories (parent directory of the "probs" directory)
    layer = "probs"
    logger.info(f"Creating necessary directories for: {loi10prod.path(layer)}")
    loi10prod.path(layer).parent.mkdir(parents=True, exist_ok=True, mode=0o775)
    logger.info(
        f"LOI product created successfully. Path: {loi10prod.path(layer)}"
    )

    return loi10prod


def create_loi10_product_aws(
    s2_product_id: str,
    project_name: str,
    version: str,
    products_base_path: Path,
    product_name: str,
) -> Loi10Product:
    """
    Create a Loi10Product object and the necessary directories using AWS resources.
    """

    loi10prod = Loi10Product(
        s2_product_id=s2_product_id,
        project_name=project_name,
        version=version,
        products_base_path=str(products_base_path),
        product_name=product_name,
    )

    if not (
        loi10prod.products_base_path.as_posix().startswith("s3://")
        or loi10prod.products_base_path.as_posix().startswith("/vsis3/")
    ):
        loi10prod.path("probs").parent.mkdir(
            parents=True, exist_ok=True, mode=0o775
        )
    return loi10prod


def extract_product_info(
    product_collection: object, s2grid: object
) -> Tuple[int, tuple, str]:
    """
    Extract the EPSG code, bounds, and date from the product collection.

    Args:
        product_collection (object): Product collection object from veg_collections.
        s2grid (object): S2Grid object from veg_collections.

    Returns:
        Tuple[int, tuple, str]: A tuple containing the EPSG code, bounds, and date.
    """
    product_tile = product_collection.df.tile.unique()[0]
    epsg = s2grid[s2grid.tile == product_tile].epsg.values[0]
    # Note: .bounds != ["bounds"]
    bounds = s2grid[s2grid.tile == product_tile]["bounds"].values[0]
    date = product_collection.df.date.values[0]
    logger.info(
        f"Product information extracted successfully: EPSG={epsg}, bounds={bounds}, date={date}, tile={product_tile}"
    )
    return bounds, date, epsg, product_tile


def filter_product_collection(
    veg_collection: object, s2grid: object, s2_product_id: str
) -> object:
    """
    Filter the product collection based on the Sentinel-2 product ID.

    Args:
        veg_collection (object): Collection object from veg_collections.
        s2grid (object): S2Grid object from veg_collections.
        s2_product_id (str): Sentinel-2 product ID.

    Returns:
        object: Filtered product collection object.
    """
    product_collection = veg_collection.__class__(
        veg_collection.df[veg_collection.df.product_id == s2_product_id],
        s2grid,
    )
    assert len(product_collection.df), f"No products found for {s2_product_id}"
    return product_collection


def get_collection_and_s2grid(collection_name: str) -> Tuple[object, object]:
    """
    Load the collection and s2grid from the veg_collections module.

    Args:
        collection_name (str): Name of the collection from veg_collections (e.g. "lcfm_1percent", "lcfm_10percent").

    Returns:
        Tuple[object, object]: A tuple containing the collection and s2grid objects.
    """
    collection = getattr(veg_collections, collection_name)
    s2grid = veg_collections.s2grid
    logger.info("Collection and S2 grid loaded successfully")
    return collection, s2grid


if __name__ == "__main__":
    # Initialize the pipeline
    cloud_inferencer = CloudInferencePipeline(
        model_name="2501241011_v8aa",
        model_idx=2,
        version="v999",
        export_products=False,  # Set to True if you want to save GeoTIFF outputs
    )

    # Get input data
    input_array, metadata = cloud_inferencer.get_input(
        product_id="S2A_MSIL2A_20200612T023601_N0500_R089_T50NKJ_20230327T190018",
        collection_name="lcfm_10percent",
    )
    # Preprocess input data
    prepared_input_array, feats_head = cloud_inferencer.preprocess(input_array)

    # Run inference
    prob_output, class_output = cloud_inferencer.run_inference(
        prepared_input_array, feats_head, export=True
    )
